function trace_str = fun_graph_trace_along_link_from_a_node(vessel_graph, link_label, node_label, tracing_order, min_radius)
% fun_grpah_trace_along_link_from_a_node find the downstream links and
% nodes label for a given link, start from one of its nodes. 
% Input: 
%   vessel_graph: structure generated by fun_skeleton_to_graph and
%   fun_graph_add_radius;
%   link_label: numerical scalar, label of the link in the vessel graph
%   node_label: numerical scalar, label of the node in the vessel graph
%   tracing_order: the maximum order of downstream links to trace
%   min_radius: if not 0, only trace the links that have median radius
%   greater than this threshold
% Output: 
%   trace_str: structure with fields: 
%       link_label_at_level: cell array, the i-th cell contains the
%       downstream link label at order i;
%       node_label_at_level: cell array, the i-th cell contains the
%       downstream node label at order i;
% 
% Implemented by Xiang Ji on 07/14/2019

if nargin < 5
    min_radius = 0;
end
assert(tracing_order >= 1, 'Tracing order should be not smaller than 1');
trace_str = struct;
if isfinite(tracing_order)
    [trace_str.link_label_at_level, trace_str.node_label_at_level] = deal(cell(tracing_order, 1));
end
% Set the current link as level 1...
trace_str.node_label_at_level{1} = node_label;
first_order_link_label = vessel_graph.node.connected_link_label{node_label};
assert(any(first_order_link_label == link_label), sprintf('Node %d is not a node of link %d\n', ...
    node_label, link_label));
if min_radius > 0
    med_radius = cellfun(@(x) median(full(vessel_graph.radius(x))), vessel_graph.link.cc_ind(first_order_link_label));
    first_order_link_label = first_order_link_label(med_radius > min_radius);
end
first_order_link_label = first_order_link_label(first_order_link_label ~= link_label);
trace_str.link_label_at_level{1} = first_order_link_label;

current_level = 2;
while current_level <= tracing_order
    connected_node_label = vessel_graph.link.connected_node_label(trace_str.link_label_at_level{current_level - 1}, :);
    connected_node_label = connected_node_label(connected_node_label > 0);
    connected_node_label = setdiff(connected_node_label, cat(1, trace_str.node_label_at_level{1 : (current_level - 1)}), 'stable');
    
    connected_link_label = cat(1, vessel_graph.node.connected_link_label{connected_node_label});
    connected_link_label = setdiff(connected_link_label, cat(1, trace_str.link_label_at_level{1 : (current_level - 1)}), 'stable');
    if min_radius > 0
        med_radius = cellfun(@(x) median(full(vessel_graph.radius(x))), vessel_graph.link.cc_ind(connected_link_label));
        connected_link_label = connected_link_label(med_radius > min_radius);
    end
    
    if isempty(connected_node_label) || isempty(connected_link_label)
        break;
    end
    trace_str.node_label_at_level{current_level} = connected_node_label;
    trace_str.link_label_at_level{current_level} = connected_link_label;
    current_level = current_level + 1;
end
trace_str.link_label = cat(1, trace_str.link_label_at_level{:});
trace_str.node_label = cat(1, trace_str.node_label_at_level{:});
trace_str.max_level = numel(trace_str.node_label_at_level);
trace_str.link_order = repelem(1:trace_str.max_level, cellfun(@numel, trace_str.link_label_at_level)).';
trace_str.node_order = repelem(1:trace_str.max_level, cellfun(@numel, trace_str.node_label_at_level)).';
end